const { ImportMethods, UIDProvider } = require( 'o-toolbox' )
const EventEmitter = require( 'events' )
const BluePrintCommon = require( './BluePrintCommon' )
const BrowserDevice = require( './BrowserDevice' )
const uidProvider = new UIDProvider({ salt: 'BluePrint' })

class ProgramBluePrint extends EventEmitter {
  static createUID () {
    return uidProvider.createUID()
  }

  constructor (params) {
    super()
    this.isInvalidated = false
    this.browserDevice = params
                           ? params.browserDevice
                           : undefined
    this.router = params
                    ? params.router
                    : undefined
    this.publicCommands = [
      'createUID',
      'evaluate',
      'getBrowserDevice',
      'getCurrentPath',
      'getRouter',
      'getUID',
      'gotoPath',
      'gotoPathLocally',
      'invalidate',
      'invalidateNow',
      'on',
      'onInvalidate',
      'respondsTo',
      'route',
      'routeWithMethod',
      'setBrowserDevice',
      'setPath'
    ]
    this.uid = this.createUID()
    if ( !this.browserDevice ) {
      this.useWeBrowserDevice()
    }
    this.initialize( params )

    return this.asReference()
  }

  acceptsCommand ({ command }) {
    return this.publicCommands.includes( command )
  }

  createUID () {
    return ProgramBluePrint.createUID()
  }

  evaluate ({ command, params }) {
    return this.acceptsCommand({ command })
             ? Reflect.apply(
                 this[ command ],
                 this,
                 params
               )
             : this.raiseCommandNotFoundError({ command, params })
  }

  getBrowserDevice () {
    return this.browserDevice
  }

  getClassName () {
    return this.constructor.name
  }

  getCurrentPath () {
    return this.browserDevice.getCurrentPath()
  }

  getRouter () {
    return this.router
  }

  getUID () {
    return this.uid
  }

  gotoPath ({ path }) {
    this.browserDevice.gotoPath({ path })
  }

  gotoPathLocally ({ path }) {
    this.browserDevice.gotoPathLocally({ path })
  }

  invalidate () {
    if ( this.isInvalidated === true ) { return }
    this.isInvalidated = true
    process.nextTick( () => this.invalidateNow() )
  }

  invalidateNow () {
    this.isInvalidated = false
    this.emit( 'invalidate' )
  }

  onInvalidate (block) {
    this.on( 'invalidate', block )
  }

  respondsTo ({ command }) {
    return this.acceptsCommand({ command })
  }

  route () {
    return this.getRouter()
             .route({ bluePrint: this })
  }

  routeWithMethod ({ methodName }) {
    return this.getRouter()
             .routeWithMethod({ methodName, bluePrint: this })
  }

  setBrowserDevice ({ browserDevice }) {
    this.browserDevice = browserDevice
  }

  setPath ({ path }) {
    this.browserDevice.setPath({ path })
  }

  useWeBrowserDevice () {
    const browserDevice = new BrowserDevice()
    this.setBrowserDevice({ browserDevice })
  }
}
ImportMethods.all({ from: BluePrintCommon, into: ProgramBluePrint })

module.exports = ProgramBluePrint